/* ----------------------------------------------------------------------

     filename: colour.c		program: colour		project: qcd

   ----------------------------------------------------------------------

     this program calculates the colour factor of qcd diagrams. Some
     restrictions apply; no external legs are allowed, maximum 200 gluons
     can be included, no four gluon vertex are allowed directly. 

     for a longer description of the program and how to use it you have
     to read the preprint

	J. H\"akkinen, H. Kharaziha, LU-TP 96-8, February 1996.

     The article is also submitted to Computer Physics Communications.
     

     this program was written in early 1995 by
	Jari H\"akkinen,	jari@thep.lu.se
	Hamid Kharraziha,	hamid@thep.lu.se

	Department of Theoretical Physics 2, Lund University, Sweden 


     references to rules are done in the comments below. the equation
     numbers refer to equations in appendix b of the preprint above.

   ---------------------------------------------------------------------- */

#include "colourdfns.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <alloca.h>



void changeindex(struct term *dia, int i, int j)

{ int		k;
  struct num	*p;
  struct trace	*t;
  struct triple	*f;

  /*looking for the gluons to be updated in the antisymmetric triple vertices.*/
  f=dia->f;
  while (f) {
    for (k=0;k<3;k++)
      if (f->gluon[k]==i) {
	f->gluon[k]=j;		/* using index j for the "new" gluon. */
	return;			/* no need to find the second gluon. */
      }
      else
	if (f->gluon[k]==j) {
	  f->gluon[k]=i;	/* using index i for the "new" gluon. */
	  return;		/* no need to find the second gluon. */
	}
    f=f->next;
  }
  /* looking for the gluons to be updated in the symmetric triple vertices. */
  f=dia->d;
  while (f) {
    for (k=0;k<3;k++) 
      if (f->gluon[k]==i) {
	f->gluon[k]=j;		/* using index j for the "new" gluon. */
	return;			/* no need to find the second gluon. */
      }
      else
	if (f->gluon[k]==j) {
	  f->gluon[k]=i;	/* using index i for the "new" gluon. */
	  return;		/* no need to find the second gluon. */
	}
    f=f->next;
  }
  /* looking for the gluons to be updated in the fermion loops. */
  t=dia->t;
  while (t) {
    p=t->gluon;
    while (p) {
      if (p->i==i) {
	p->i=j;			/* using index j for the "new" gluon. */
	return;			/* no need to find the second gluon. */
      }
      else
	if (p->i==j) {
	  p->i=i;		/* using index j for the "new" gluon. */
	  return;		/* no need to find the second gluon. */
	}
      p=p->next;
    }
    t=t->next;
  }
} /* of changeindex */



struct num *changetrace(struct num *p1,struct trace *t1)

{ int		hupp, i, tal;
  struct num	*p2, *p3, *p4, *p5;
  struct trace 	*t2, *t3;
  
  /* this function is hard to describe. using this function when rule C19 is
     applied for anti-symmetric triple gluon vertices will give large 
     enhancements in performance. this function looks for gluons (in the vertex)
     connected to fermion loops. if one or more is connected to fermion loops
     smart things are done. actually we replace the gluon with rule C11 and some
     more... */
  hupp=0;
  for (i=0;i<3;i++) {
    t2=t1;
    t3=t1->next;
    while (t3) {
      p2=t3->gluon;
      while ((p2) && (p2->i!=p1->i))
	p2=p2->next;
      if (p2) {
	if (p2==t3->gluon)
	  t3->gluon=NULL;
	else {
	  p3=t3->gluon;
	  while (p3->next!=p2)
	    p3=p3->next;
	  p3->next=NULL;
	}
	p4=p1->next;
	free(p1);
	p1=p4;
	if (p1) {
	  while(p4->next)
	    p4=p4->next;
	  p4->next=p2->next;
	}
	else {
	  p1=p2->next;
	  p4=p1;
	}
	free(p2);
	if (p1)	{
	  while(p4->next)
	    p4=p4->next;
	  p4->next=t3->gluon;
	}
	else
	  p1=t3->gluon;
	t2->next=t3->next;
	free(t3);
	t3=NULL;
	hupp=1;
      }
      else {
	t2=t2->next;
	t3=t3->next;
      }
    }
    if ((i) && (!hupp))	{
      p2=p1;
      while (p2) {
	p3=p2;
	p2=p2->next;
	if ((p2) && (p1->i==p2->i)) {
	  t2=gettrace();
	  t2->next=t1->next;
	  t1->next=t2;
	  t2->gluon=p2->next;
	  p3->next=NULL;
	  free(p2);
	  p2=p1;
	  p1=p1->next;
	  free(p2);
	  hupp++;
	  p2=NULL;
	}
      }
    }
    if ((!hupp)&&(i<2))	{
      p2=p1;
      while (p2->next)
	p2=p2->next;
      p2->next=p1;
      p1=p1->next;
      p2=p2->next;
      p2->next=NULL;
    }
    else hupp=0;
  } 
  return p1;
} /* of changetrace */



int checktrace(struct term *dia)

{ int		i, j;
  struct num	*p1, *p2;
  
  /* trying to enforce rules x-x on the fermion loop. */
  if (!dia->t)				/* there is no fermion loop. */
    return 0;
  p1=dia->t->gluon;
  if (!p1)				/* there is no gluons attached to the */
    return 0;				/* fermion loop. */
  if (!p1->next) {			/* applying rule C2. zero contribution*/
    killterm(dia);			/* from this term. */
    return 1;
  } 
  p2=p1->next;
  if (p1->i==p2->i) {			/* applying rule C14 */
    dia->t->gluon=p2->next;
    free(p1);
    free(p2);
    dia->ecf++;
    return 1;
  }
  if (!p2->next) {			/* applying rule C18 */
    i=p1->i;
    j=p2->i;
    free(p1);
    free(p2);
    dia->t->gluon=NULL;			/* should remove fermion loop but */
    dia->e--;				/* compensating with this factor. */
    /* gluon indexes must be updated, "two gluons goes to one gluon" */
    changeindex(dia,i,j);
    return 1;
  }
  p2=p2->next;
  /* looking for rule C15 or C14 cases */
  while ((p2) && (p2->i!=p1->i) && (p1->next->i!=p2->i)) {
    p2=p2->next;
    p1=p1->next;
  }
  if (p2) { 
    if (p1->next->i==p2->i) {		/* applying rule C14 */
      free(p1->next);
      p1->next=p2->next;
      free(p2);
      dia->ecf++;
      return 1;
    }
    /* from now we know that rule C15 should be applied. */
    if (p1==dia->t->gluon) {
      dia->t->gluon=p1->next;
      free(p1);
      p2=dia->t->gluon;
    }
    else {
      p2=dia->t->gluon;
      while (p1!=p2->next)
	p2=p2->next;
      p2->next=p1->next;
      free(p1);
      p2=p2->next;
    }
    p1=p2->next;
    p2->next=p1->next;
    free(p1);
    dia->a*= -1;
    dia->e--;
    return 1;
  }
  return 0;
} /* of checktrace */



void copyterm(struct term *dia)

{ int 		i;
  struct term	*dianew;
  struct trace	*t1, *t2, *t3;
  struct triple *f1, *f2, *f3;
  struct num	*p1, *p2, *p3;

  /* creating another term and making a carbon copy of the first item in the
     (dia) list. */
  dianew=getterm();
  dianew->next=dia->next;
  dia->next=dianew;
  dianew->a=dia->a;
  dianew->e=dia->e;
  dianew->ecf=dia->ecf;
  t1=gettrace();
  t3=t1;
  t2=dia->t;
  p1=getnum();
  while (t2) {
    t3->next=gettrace();
    t3=t3->next;
    p1->next=NULL;
    p3=p1;
    p2=t2->gluon;
    while(p2) {
      p3->next=getnum();
      p3=p3->next;
      p3->i=p2->i;
      p2=p2->next;
    }
    t3->gluon=p1->next;
    t2=t2->next;
  }
  dianew->t=t1->next;
  free(p1);
  free(t1);
  f1=gettriple();
  f3=f1;
  f2=dia->f;
  while(f2) {
    f3->next=gettriple();
    f3=f3->next;
    for (i=0;i<3;i++)
      f3->gluon[i]=f2->gluon[i];
    f2=f2->next;
  }
  dianew->f=f1->next;
  f1->next=NULL;
  f3=f1;
  f2=dia->d;
  while(f2) {
    f3->next=gettriple();
    f3=f3->next;
    for (i=0;i<3;i++)
      f3->gluon[i]=f2->gluon[i];
    f2=f2->next;
  }
  dianew->d=f1->next;
  free(f1);
} /* of copyterm */



void error(int errno)	/* no error handling, prints a message and exits */

{ 
  switch(errno) {
  case 0:
    printf("Allocation error - out of memory\n");
    printf("This error is severe. Either there is major problems\n");
    printf("with the computer memory or this diagram is to big to\n");
    printf("digest. Please contact your system manager if the\n");
    printf("second problem seems unprobable\n");
    break;
  case 1:
    printf("Syntax error in the input file.\n");
    break;
  case 2:
    printf("Syntax error in the input file.\n");
    printf("One gluon number occurs only once.\n"); 
    break;
  case 3:
    printf("Syntax error in the input file.\n");
    printf("One gluon number occurs more than twice.\n"); 
    break;
  case 4:
    printf("Underflow occured, power of N less than %d.\n",
	   (int)(-MAXPOW/2));
    break;
  case 5:
    printf("Overflow occured, power of N greater than %d.\n",
	   (int)(MAXPOW/2-1));
    break;
  } /* of switch */
  exit(-1);
} /* of error */



int expand(struct term **p,int n[MAXPOW]) 

{ int		differ;
  struct term	*dia, *diatmp;
  struct trace	*t1;

  dia= *p;
  if (!dia)				/* null pointer, diagram finished. */
    return 0;
  /* the performance of the program depends on which order different
     diagrammatic rules are used. the order below seems to be fairly fast. */
  if (checktrace(dia))		/* starting by manipulating fermion loops. */
    return 1;
  if (dia->f) {			/* treat anti symmetric triple vertex. */
    triplevertex(dia,1);
    return 1;
  }
  if (dia->d) {			/* treat symmetric triple vertex. */
    triplevertex(dia,0);
    return 1;
  }
  if (!dia->t) {		/* no fermion loops, term finished. */
    inputn(dia,n);		/* update n array. */
    diatmp=dia;
    *p=diatmp->next;		/* term finished, move to next term. */
    free(diatmp);
    return 1;
  }
  /* now only fermion loops are left in this term. */
  if (!dia->t->gluon) {
    dia->e++;			/* empty loop, gives an overall N_c to term. */
    t1=dia->t;
    dia->t=t1->next;		/* next fermion loop. */
    free(t1);
    return 1;
  }
  /* fermion loop with gluons, rule C11. */
  copyterm(dia);		/* create a carbon copy in list. */
  singlet(dia->next);		/* replace gluon with singlet. */ 
  whole(dia);			/* replace gluon with a 3x3~ state. */
  return 1;
} /* of expand */


struct num *getnum()

{ struct num	*p;

  if ((p=(struct num *)malloc(sizeof(struct num))) == NULL)
    error(0);				/* allocating memory. */
  p->i=0;				/* initializing the structure. */
  p->next=NULL;
  return p;
} /* of getnum. */



struct term *getterm()

{ struct term	*p;

  if ((p=(struct term *)malloc(sizeof(struct term))) == NULL)
    error(0);				/* allocating memory. */
  p->a=1;				/* initializing the structure. */
  p->e=0;
  p->ecf=0;
  p->d=NULL;
  p->f=NULL;
  p->t=NULL;
  p->next=NULL;
  return p;
} /* of getterm. */



struct trace *gettrace()

{ struct trace	*p;

  if ((p=(struct trace *)malloc(sizeof(struct trace))) == NULL)
    error(0);				/* allocating memory. */
  p->gluon=NULL;			/* initializing the structure. */
  p->next=NULL;
  return p;
} /* of gettrace. */



struct triple *gettriple()

{ struct triple	*p;

  if ((p=(struct triple *)malloc(sizeof(struct triple))) == NULL)
    error(0);				/* allocating memory. */
  p->gluon[0]=0;			/* initializing the structure. */
  p->gluon[1]=0;
  p->gluon[2]=0;
  p->next=NULL;
  return p;
} /* of gettriple. */



void init(argc,argv,infilename,n)
	char	*argv[];
	char	infilename[];
	int	argc, n[MAXPOW];
	
{ char	ch;
  FILE	*helpfile;
  int 	i, ok;

  if (((int)(MAXPOW)/2-(double)(MAXPOW)/2)<-0.25) {
    printf("MAXPOW is not defined as an even number\n"); 
    printf("Please make the correction and recompile\n"); 
    exit(-1);
  }
  for (i=0;i<MAXPOW;i++) 
    n[i]=0;				/* resetting n polynomial vector. */
  strcpy(infilename,DFLTINFILE);	/* default input filename. */
  /* analysing parameters. */
  for (i=1;i<argc;i++)			/* checking for help request. */
    if ((!strcmp(argv[i],"-help")) || (!strcmp(argv[i],"-h"))) {
      printf("\n\nProgram: colour         Version 1.0         Last update: February 29, 1996\n");
      printf("\n\nThis program calculates colour factors for QCD diagrams. A longer\n");
      printf("description of the program is presented in the preprint LU TP 96-8 written\n");
      printf("by Jari H\\\"akkinen and Hamid Kharraziha, Department of Theoretical Physics,\n");
      printf("Lund University, Sweden\n"); 
      printf("\n\nSyntax:           colour [-df file | -h | -help]\n\n");
      printf("    -df           Changes the inputfile name. The default is colour.in\n");
      printf("    -h or -help   Prints this page.\n");
      printf("\n\nAll input to the program is done through an input file. The rules for how\n");
      printf("to define a qcd diagram in the file are:\n\n"); 
      printf("    * Every gluon line is assigned a unique number.\n"); 
      printf("    * Triple anti-symmetric gluon vertices are represented by a letter f\n");
      printf("      followed by the three incoming gluon numbers.\n");
      printf("    * Symmetric triple vertices (are not physical QCD gluon vertices, but\n");
      printf("      are included for generality) are represented in the same way as the\n");
      printf("      anti-symmetric vertices but with the letter d.\n");
      printf("    * Quark loops are represented by a letter t, followed by all gluons\n");
      printf("      connected to the quark loop.\n");
      printf("    * Four-gluon vertices cannot be directly included in the program since\n");
      printf("      the colour and the kinematical parts of the cross section do not\n");
      printf("      factorize. This can, however, be taken care of by separating the\n");
      printf("      four-gluon vertex in factorizable terms, as discussed further in\n"); 
      printf("      in the LU TP 96-8 preprint\n"); 
      printf("    * Ghosts can also be included since these couple to gluons with the\n"); 
      printf("      same colour factor as gluons themselves. Thus ghost vertices can be\n");
      printf("      represented with a letter f.\n");
      printf("    * For the quark loops and the anti-symmetric gluon vertex the order of\n"); 
      printf("      the gluons are important since f_{abc} = -f_{bac}. So when translating\n");
      printf("      a diagram into a file all loops and f vertices must be written down\n");
      printf("      in the same order (clockwise or anti-clockwise).\n");
      printf("    * Composite objects, d, f, and t, must be separated with a newline\n");
      printf("      character, while gluon numbers are separated with a space. All\n");
      printf("      other separation characters are invalid and will stop the\n");
      printf("      execution of the program.\n\n\n");
      exit(1);                	/* always exit after printing help. */
    }
  printf("\n");
  for (i=1;i<argc;i++) {		/* checking options. */
    ok=FALSE;
    if (!strcmp(argv[i],"-df")) {
      if ((i+1)<argc)
	strncpy(infilename,argv[++i],MAXSTRING);
      if (!strlen(infilename))
	strcpy(infilename,DFLTINFILE);
      else
	printf("parameter -df given, changing input file to: %s\n",infilename);
      ok=TRUE;
    }
    if (!ok)
      printf("invalid option: %s\n",argv[i]);
  }
} /* of init. */



void inputn(struct term *dia,int n[])

{ int	a, e, ecf, i;
  
  a=dia->a;
  e=dia->e;
  ecf=dia->ecf;
  for (i=0;i<=ecf;i++) {
    if ((int)(MAXPOW/2+e+ecf-2*i)<0)
      error(4);			/* underflow. */
    else
      if ((int)(MAXPOW/2+e+ecf-2*i)>(MAXPOW-1))
	error(5);		/* overflow. */
      else
	n[(int)(MAXPOW/2)+e+ecf-2*i]+=a;
				/* update n array, n[MAXPOW/2] --> m*N_c^0 */
    a*=(i-ecf);			/* C_F = (N^2-1)/(2N) */
    a=a/(i+1);
  }
}/* of inputn */



void killterm(struct term *dia)

{ struct num	*p1, *p2;
  struct trace	*t1, *t2;
  struct triple *f1, *f2;

  /* removing the term pointed at by *dia. */
  dia->a=0;
  f1=dia->f;
  while (f1) {
    f2=f1;
    f1=f1->next;
    free(f2);
  }
  dia->f=NULL;
  f1=dia->d;
  while (f1) {
    f2=f1;
    f1=f1->next;
    free(f2);
  }
  dia->d=NULL;
  t1=dia->t;
  while (t1) {
    t2=t1;
    p1=t1->gluon;
    while (p1) {
      p2=p1;
      p1=p1->next;
      free(p2);
    }
    t1=t1->next;
    free(t2);
  }
  dia->t=NULL;
} /* of killterm */



void main(int argc,char *argv[])

{ char		infilename[MAXSTRING];
  /* n represents the result in powers of N_c. read as
     n[0]*N^(0-MAXPOW/2)+n[1]*N^(1-MAXPOW/2)+...+
			n[i]*N^(i-MAXPOW/2)+...+n[199]*N^(MAXPOW-1-MAXPOW/2) */
  int		n[MAXPOW];
  int		normexp;		/* overall norm. factor, 2^normexp */
  struct term	*dia;			/* list of terms in the diagram. */

  init(argc,argv,infilename,n);		/* initializing, reading options. */
  dia=readfile(infilename,&normexp);	/* reading inputfile,
					   setting up pointer structure. */
  while(expand(&dia,n));		/* calculating the diagram */
  writeoutput(normexp,n);		/* writing the result */
} /* of main */



struct term *readfile(infilename,normexp)
	char	infilename[];
	int	*normexp;

{ char		ch, type;
  int		i, j, more, val;
  FILE		*input;
  struct num	*gluon;
  struct term 	*dia, *dia2;
  struct trace	*t;
  struct triple	*d, *f;
  int		numcheck[2][MAXGLUONS];

  /* normexp keeps track of overall normalization. the normalization factor
     is given by
	2^(-normexp)
     where
	normexp = #triplevertices - #gluons		*/
  (*normexp)=0;			
  for (i=0;i<MAXGLUONS;i++)	/* numcheck is used to check syntax in input */
    numcheck[1][i]=0;		/*file and to calculate overall normalization.*/
  if ((input=fopen(infilename,"r"))==NULL) {
    printf("cannot open input file.\n\ntype colour -h to get help.\n\n");
    exit(-1);
  }
  more=TRUE;
  dia=getterm();		/* getting term pointer and allocating memory.*/
  ch=getc(input);		/* reading first character in input file. */
  do {
    if (ch=='d' || ch=='f' || ch=='t') {
      type=ch;			/* need to remember which type of object since*/
      switch(ch) {		/* gluons are defined after object type. */
      case 'd':			/* symmetric triple gluon vertex. */
	d=gettriple();		/* getting triple pointer, allocating memory. */
	d->next=dia->d;		/* inserting the vertex into dia. */
	dia->d=d;
	(*normexp)++;		/* keeping track of overall normalization. */
	break;
      case 'f':			/* anti-symmetric triple gluon vertex. */
	f=gettriple();		/* getting triple pointer, allocating memory. */
	f->next=dia->f;		/* inserting the vertex into dia. */
	dia->f=f;
	(*normexp)++;		/* keeping track of overall normalization. */
	break;
      case 't':			/* quark ring (trace). */
	t=gettrace();		/* getting trace pointer, allocating memory. */
	t->next=dia->t;		/* inserting the trace into dia. */
	dia->t=t;
	break;			/* no update on overall normalization. */
      }
      /* 'i' used to keep track of the number of arguments after the object
	 type. 'val' is used to convert characters to integer since the input
	 file is read as characters. */
      i=val=0;		
      ch=getc(input);		/* read next character from input file. */
    }
    /* different cases depending on the value of 'ch' and 'type'.
       the program expects three parameters after object types f and d, while
       object t can have any number of parameters.
       all gluons are numbered with integers > 0. the integer are separated
       with blank spaces.
       obvious syntax errors in the input file is detected. */
    switch(ch) {
    case ' ':	       
    case '\n':
    case EOF:
      switch(type) {
      case 'd':
	/* after a new object is read, 'val'=0 and next character should be 
	   read. */
	if (val)
	  if (i<3)			/* only three parameters are expected.*/
	    dia->d->gluon[i++]=val;	/* store gluon number */
	  else
	    error(1);			/* to many parameters detected. */
	break;
      case 'f':
	/* after a new object is read, 'val'=0 and next character should be 
	   read. */
	if (val)
	  if (i<3)			/* only three parameters are expected.*/
	    dia->f->gluon[i++]=val;	/* store gluon number */
	  else
	    error(1);			/* to many parameters detected. */
	break;
      case 't':
	/* after a new object is read, 'val'=0 and next character should be 
	   read. */
	if (val) {
	  gluon=getnum();		/* get num pointer, allocate memory. */
	  gluon->i=val;			/* store gluon number and insert into */
	  gluon->next=dia->t->gluon;  	/* trace in dia. */
	  dia->t->gluon=gluon;
	}
	break;
      default:				/* undefined type in input file. */
	error(1);
      } /* of switch */
      /* if a gluon number is read ('val'!=0), store normalizations factor
	 information for future analysis.
	 'numcheck[0][n]' contains the gluon numbers ('val'>0) encountered
	 sofar, and 'numcheck[1][n]' counts the number of occurences of
	 gluon number 'val'.				       */
      if (val) {
	j= -1; 
	while((numcheck[1][++j]!=0) && (numcheck[0][j]!=val));
	if (numcheck[1][j]==0)
	  numcheck[0][j]=val;
	numcheck[1][j]++;
      }
      val=0;				/* reset val. */
      ch=getc(input);			/* read character from input file. */
      break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':				/* read a number, updating 'val'. */
      val*=10;				/* decimal system. */
      val+=ch-48;			/* ASCII code for 0 starts a 48 */
      ch=getc(input);			/* read character from input file. */
      break;
    default:
      error(1);				/* unrecognized character. */
    } /* of switch */
    if ((ch==EOF) && (!val))
      more=FALSE;			/* end of input file reached. */
  } while (more);
  fclose(input);
  /* calculating the overall normalization and checking the numbering of the 
     gluons */
  for (j=0;j<MAXGLUONS;j++)
    switch(numcheck[1][j]) {
    case 0:
      /* all gluons treated */
      j=MAXGLUONS;
      break;
    case 1:
      /* gluon is referenced only once, every gluon in the diagram must be
	 referenced two (2) times, no more or less */
      error(2);		
      break;
    case 2:
      /* every gluon multiply an overall factor 1/2 to the diagram */
      (*normexp)--;
      break;
    default:
      /* gluon is referenced more then twice, every gluon in the diagram must
	 be referenced two (2) times, no more or less */
      error(3);
    } /* of switch */
  return dia;
} /* of readfile */



void singlet(struct term *dia)
 
{ int		i, match;
  struct trace	*t, *t2;
  struct num	*p, *p2;

  /* replacing a gluon with a singlet, needed for rule C11. */
  match=FALSE;
  dia->a*= -1;
  dia->e--;
  t=dia->t;
  i=t->gluon->i;
  p=t->gluon->next;
  free(t->gluon);
  t->gluon=p;  
  while(!match) {
    p2=p;
    while((p) && (p->i!=i)) {
      p2=p;
      p=p->next;
    }
    t2=t;
    t=t->next;
    if ((!p) && (!t))
      error(2);
    else
      if (p)
	match=TRUE;
      else
	p=t->gluon;
  }
  if (p2==p)
    t2->gluon=p->next;
  else
    p2->next=p->next;
  free(p);
} /* of singlet */



void triplevertex(struct term *dia,int flag)

{ int 		i, j,gluon[3];
  struct num	*p1, *p2,*p3;
  struct trace	*t1, *t2;
  struct triple *f1;
  
  /* flag tells if a symmetric or anti symmetric triple vertex is to be 
     treated. */
  /* extracting gluon indices from the triple vertex. */
  if (flag) {
    for (i=0;i<3;i++) 
      gluon[i]=dia->f->gluon[i];
    f1=dia->f->next;
    free(dia->f);			/* removing the vertex. */
    dia->f=f1;
  }
  else {
    for (i=0;i<3;i++) 
      gluon[i]=dia->d->gluon[i];
    f1=dia->d->next;
    free(dia->d);			/* removing the vertex. */
    dia->d=f1;
  }
  if ((gluon[0]==gluon[1]) || (gluon[0]==gluon[2]) || (gluon[1]==gluon[2])) {
    killterm(dia);			/* applying rule C7, zero contribution*/
    return;				/* from this term. */
  }
  /* a triple vertex expands to two terms, see rule C19 (modified). */
  copyterm(dia);
  t1=gettrace();			/* the first fermion loop. */
  t1->next=dia->t;
  dia->t=t1;
  /* p1 is one of the new fermion loops. p2 is running. */
  p1=p2=getnum();		
  for (i=0;i<3;i++) {		/* order of the gluons connected to the loop. */
    p2->i=gluon[i];
    if (i!=2) {
      p2->next=getnum();
      p2=p2->next;
    }
  }
  if ((flag) && (t1->next))
    p1=changetrace(p1,t1);	/* make performance enhancements. */
  t1->gluon=p1;
  t1=gettrace();		/* the second fermion loop. */
  t1->next=dia->next->t;
  dia->next->t=t1;
  /* p1 is now the other one of the fermion loops. p2 is still running */
  p1=p2=getnum();
  for (i=2;i>-1;i--) {		/* order of the gluons connected to the loop. */
    p2->i=gluon[i];
    if (i) {
      p2->next=getnum();
      p2=p2->next;
    }
  }
  if ((flag) && (t1->next))
    p1=changetrace(p1,t1);	/* make performance enhancements. */
  t1->gluon=p1;
  if (flag)
    dia->next->a*= -1;	/* sign is different for the two triple vertices. */
} /* of triple */



void whole(struct term *dia)
/* must come after checktrace */

{ struct trace	*t1, *t2, *t3;
  struct num	*p1, *p2, *p3;
  
  /* changing a gluon to a 3x3~ state, neede for rule C11. */
  t1=dia->t;
  p1=t1->gluon;
  p2=p1->next;
  while ((p2->next) && (p2->next->i!=p1->i))
    p2=p2->next;
  if (p2->next) {
    t1->gluon=p1->next;
    free(p1);
    t2=gettrace();
    t2->next=t1->next;
    t1->next=t2;
    t2->gluon=p2->next->next;
    free(p2->next);
    p2->next=NULL;
    return;
  }
  t2=t1->next;
  while (t2) {
    if (t2->gluon) {
      p2=t2->gluon;
      if (p2->i==p1->i) {
	t2->gluon=p2->next;
	free(p2);
	if (t2->gluon) {
	  p2=t2->gluon;
	  while(p2->next)
	    p2=p2->next;
	  p2->next=p1->next;
	}
	else {
	  t2->gluon=p1->next;
	  p2=p1;
	}
      }
      else {
	while ((p2->next) && (p2->next->i!=p1->i))
	  p2=p2->next;
	if (p2->next) {
	  p3=p2->next;
	  p2->next=p1->next;
	  while (p1->next)
	    p1=p1->next;
	  p1->next=p3->next;
	  free(p3);
	  p1=t1->gluon;
	}
      }
      if (p2->next) {
	free(p1);
	dia->t=t1->next;
	free(t1);
	return;
      }
    }
    t2=t2->next;
  }
  error(2);
}/* of whole */



void writeoutput(normexp,n)
	int	normexp;
	int	n[MAXPOW];

{ int i;
  double sum=0.;

  printf("\nThe answer is:\n\n"); 
  for (i=0;i<MAXPOW;i++)
    if (n[i]) {
      printf("%d N^%d\n",n[i],i-MAXPOW/2); 
      sum+=n[i]*pow(3.,(double)(i-MAXPOW/2));
    }
  printf("...times an overall factor: 2^%d, if we use ",normexp);
  printf("Gell-Mann normalization of gluons.\n");
  printf("\nFor N=3: %g using Gell-Mann normalization.\n\n",
	 sum*pow(2.,(double)(normexp))); 
}



